# ---------------------------------------------------------------------
# Gufo Ping: PingSocket test
# ---------------------------------------------------------------------
# Copyright (C) 2022-23, Gufo Labs
# ---------------------------------------------------------------------

# Python modules
import asyncio

# Third-party modules
import pytest

# Gufo Labs modules
from gufo.ping.socket import PingSocket

from .util import caps


@pytest.mark.skipif(caps.is_denied, reason="Permission denied")
@pytest.mark.parametrize(
    ("afi", "expected"), [(1, False), (4, True), (6, True)]
)
def test_afi(afi: int, expected: bool) -> None:
    async def inner_fail() -> None:
        with pytest.raises(ValueError):
            PingSocket(afi=afi)

    async def inner_ok() -> None:
        PingSocket(afi=afi)

    asyncio.run(inner_ok() if expected else inner_fail())


@pytest.mark.skipif(caps.is_denied, reason="Permission denied")
def test_empty_read() -> None:
    async def inner() -> None:
        s = PingSocket(afi=4)
        r = s._on_read()
        assert r is None

    asyncio.run(inner())


@pytest.mark.skipif(caps.is_denied, reason="Permission denied")
@pytest.mark.parametrize(
    ("afi", "addr", "expected"),
    [
        #  IPv4
        (4, "127.0.0.1", "127.0.0.1"),
        (4, "127.0.1", None),
        (4, "127.0.0.01", None),
        # IPv6
        (6, "::1", "::1"),
        (6, "0::1", "::1"),
    ],
)
def test_clean_ip(afi: int, addr: str, expected: str) -> None:
    async def inner_ok() -> None:
        s = PingSocket(afi=afi)
        assert s.clean_ip(addr) == expected

    async def inner_fail() -> None:
        s = PingSocket(afi=afi)
        with pytest.raises(ValueError):
            s.clean_ip(addr)

    asyncio.run(inner_ok() if expected else inner_fail())



# ---------------------------------------------------------------------
# Gufo Ping: Test Ping
# ---------------------------------------------------------------------
# Copyright (C) 2022-23, Gufo Labs
# ---------------------------------------------------------------------

# Python modules
import asyncio
from typing import Any, Dict, Iterable, List, Optional, Union

# Third-party modules
import pytest

# Gufo Labs modules
from gufo.ping import Ping

from .util import as_str, caps


@pytest.mark.parametrize(
    ("address", "expected"), [("127.0.0.1", 4), ("::1", 6)]
)
def test_get_afi(address: str, expected: int) -> None:
    afi = Ping._get_afi(address)
    assert afi == expected


@pytest.mark.skipif(caps.is_denied, reason="Permission denied")
@pytest.mark.parametrize(
    ("address", "expected"),
    [
        # IPv4
        ("127.0.0.1", True),  # Loopback, always available
        ("192.0.2.1", False),  # RFC-5737 test range, should fail
    ],
)
def test_ping(address: str, expected: bool) -> None:
    ping = Ping()
    rtt = asyncio.run(ping.ping(address))
    if expected:
        assert isinstance(rtt, float)
        assert rtt > 0.0
    else:
        assert rtt is None


@pytest.mark.skipif(caps.is_denied, reason="Permission denied")
@pytest.mark.parametrize(
    ("address", "expected"),
    [
        # IPv4
        ("127.0.0.1", True),  # Loopback, always available
        ("192.0.2.1", False),  # RFC-5737 test range, should fail
    ],
)
def test_iter_rtt(address: str, expected: bool) -> None:
    async def inner() -> List[Optional[float]]:
        r: List[Optional[float]] = []
        async for rtt in ping.iter_rtt(address, count=N_PROBES):
            r.append(rtt)
        return r

    N_PROBES = 5
    ping = Ping()
    res = asyncio.run(inner())
    assert len(res) == N_PROBES
    if expected:
        nr = sum(1 for rtt in res if rtt is not None)
        assert nr == N_PROBES
    else:
        nr = sum(1 for rtt in res if rtt is None)
        assert nr == N_PROBES


@pytest.mark.skipif(caps.is_denied, reason="Permission denied")
@pytest.mark.parametrize("addr", caps.loopbacks)
@pytest.mark.parametrize(
    ("cfg", "expected"),
    [
        # ttl
        ({"ttl": -1}, False),
        ({"ttl": 0}, False),
        ({"ttl": 1}, True),
        ({"ttl": 64}, True),
        ({"ttl": 255}, True),
        ({"ttl": 256}, False),
        # tos
        ({"tos": -1}, False),
        ({"tos": 0}, True),
        ({"tos": 1}, True),
        ({"tos": 64}, True),
        ({"tos": 255}, True),
        ({"tos": 256}, False),
        # send_buffer_size
        ({"send_buffer_size": 1048576}, True),
        # recv_buffer_size
        ({"recv_buffer_size": 1048576}, True),
        # coarse
        ({"coarse": True}, True),
        ({"coarse": False}, True),
    ],
    ids=as_str,
)
def test_valid_ping_settings(
    addr: str, cfg: Dict[str, Any], expected: bool
) -> None:
    if expected:
        asyncio.run(Ping(**cfg).ping(addr))
    else:
        with pytest.raises(ValueError):
            asyncio.run(Ping(**cfg).ping(addr))


@pytest.mark.parametrize(
    ("src_addr", "expected"),
    [
        (None, {}),
        ("127.0.0.1", {4: "127.0.0.1"}),
        ("::1", {6: "::1"}),
        ((), {}),
        ([], {}),
        (("127.0.0.1", "::1"), {4: "127.0.0.1", 6: "::1"}),
        (["127.0.0.1", "::1"], {4: "127.0.0.1", 6: "::1"}),
        (("::1", "127.0.0.1"), {4: "127.0.0.1", 6: "::1"}),
        (("127.0.0.1", "127.0.0.2"), {4: "127.0.0.1"}),
        (("127.0.0.2", "127.0.0.1", "::1", "::2"), {4: "127.0.0.2", 6: "::1"}),
    ],
    ids=as_str,
)
def test_src_addr(
    src_addr: Union[None, str, Iterable[str]], expected: Dict[int, str]
) -> None:
    assert Ping._get_src_addr(src_addr) == expected


@pytest.mark.skipif(caps.is_denied, reason="Permission denied")
@pytest.mark.parametrize("addr", caps.loopbacks)
def test_src_ping(addr: str) -> None:
    ping = Ping(src_addr=addr)
    rtt = asyncio.run(ping.ping(addr))
    assert isinstance(rtt, float)
    assert rtt > 0.0



# ---------------------------------------------------------------------
# Gufo Ping: _fast stub
# This is a stub file for pyo3 module only to satisfy
# the mkdocstings
# ---------------------------------------------------------------------
# Copyright (C) 2022-23, Gufo Labs
# ---------------------------------------------------------------------

# Python modules
from typing import Dict, List, Optional

class SocketWrapper(object):
    def __init__(self: "SocketWrapper", afi: int) -> None: ...
    def set_timeout(self: "SocketWrapper", timeout: int) -> None: ...
    def bind(self: "SocketWrapper", addr: str) -> None: ...
    def set_ttl(self: "SocketWrapper", ttl: int) -> None: ...
    def set_unicast_hops(self: "SocketWrapper", ttl: int) -> None: ...
    def set_tos(self: "SocketWrapper", tos: int) -> None: ...
    def set_tclass(self: "SocketWrapper", tos: int) -> None: ...
    def set_send_buffer_size(self: "SocketWrapper", size: int) -> None: ...
    def set_recv_buffer_size(self: "SocketWrapper", size: int) -> None: ...
    def set_coarse(self: "SocketWrapper", coarse: bool) -> None: ...
    def set_accelerated(self: "SocketWrapper", acc: bool) -> None: ...
    def get_fd(self: "SocketWrapper") -> int: ...
    def send(
        self: "SocketWrapper", addr: str, request_id: int, seq: int, size: int
    ) -> None: ...
    def recv(self: "SocketWrapper") -> Optional[Dict[str, int]]: ...
    def get_expired(self: "SocketWrapper") -> Optional[List[str]]: ...



# ---------------------------------------------------------------------
# Gufo Ping: PingSocket implementation
# ---------------------------------------------------------------------
# Copyright (C) 2022-25, Gufo Labs
# ---------------------------------------------------------------------

"""
PingSocket implementation.

Attributes:
    IPv4: IPv4 address family.
    IPv6: IPv6 address family.
"""

# Python modules
from asyncio import Future, get_running_loop, sleep
from typing import Dict, Optional, cast

# Gufo Labs modules
from ._fast import SocketWrapper
from .proto import SocketProto

NS = 1_000_000_000.0
IPv4 = 4
IPv6 = 6
MAX_TTL = 255
MAX_TOS = 255


class PingSocket(object):
    """
    Python API to Gufo Ping internals.

    Python-side ICMP requests/reply dispatcher for the given address family.
    Wraps Rust socket implementation.

    Args:
        afi: Address Family. Either 4 or 6
        size: Set outgoing packet's size, including IP header.
        src_addr: Optional source address of outgoing packets.
        ttl: Set outgoing packet's TTL.
            Use OS defaults when empty.
        tos: Set DSCP/TOS field to outgoing packets.
            Use OS defaults when empty.
        timeout: Default timeout in seconds.
        send_buffer_size: Send buffer size.
            Use OS defaults when empty.
        recv_buffer_size: Receive buffer size.
            Use OS defaults when empty.
        coarse: Use CLOCK_MONOTONIC_COARSE when set,
            fall back to CLOCK_MONOTONIC otherwise.
        accelerated: Enable platform-dependend accelerated
            socket processing.
    """

    VALID_AFI = (IPv4, IPv6)

    def __init__(
        self: "PingSocket",
        afi: int = IPv4,
        size: int = 64,
        src_addr: Optional[str] = None,
        ttl: Optional[int] = None,
        tos: Optional[int] = None,
        timeout: float = 1.0,
        send_buffer_size: Optional[int] = None,
        recv_buffer_size: Optional[int] = None,
        coarse: bool = False,
        accelerated: bool = True,
    ) -> None:
        self.__force_del = False
        if afi not in self.VALID_AFI:
            msg = f"afi must be {IPv4} or {IPv6}"
            raise ValueError(msg)
        # Check settings
        if ttl is not None and (ttl < 1 or ttl > MAX_TTL):
            msg = f"ttl must be in 0..{MAX_TTL} range"
            raise ValueError(msg)
        if tos is not None and (tos < 0 or tos > MAX_TOS):
            msg = f"tos must be in 0..{MAX_TOS} range"
            raise ValueError(msg)
        self.__size = size
        # Create and initialize wrapped socket
        self.__sock: SocketProto = cast(SocketProto, SocketWrapper(afi))
        self.__sock.set_timeout(int(timeout * NS))
        if src_addr:
            self.__sock.bind(src_addr)
        if ttl is not None:
            if afi == IPv4:
                self.__sock.set_ttl(ttl)
            elif afi == IPv6:
                self.__sock.set_unicast_hops(ttl)
        if tos is not None:
            if afi == IPv4:
                self.__sock.set_tos(tos)
            elif afi == IPv6:
                self.__sock.set_tclass(tos)
        if send_buffer_size is not None:
            self.__sock.set_send_buffer_size(send_buffer_size)
        if recv_buffer_size is not None:
            self.__sock.set_recv_buffer_size(recv_buffer_size)
        if coarse:
            self.__sock.set_coarse(True)
        if accelerated:
            self.__sock.set_accelerated(True)
        self.__timeout = timeout
        self.__sock_fd = self.__sock.get_fd()
        #  <addr>-<request id>-<seq> -> future
        self.__sessions: Dict[str, Future[Optional[float]]] = {}
        # Install response reader
        self.__force_del = True
        get_running_loop().add_reader(self.__sock_fd, self._on_read)
        # Install deadline cleaner
        self.__cleanup_task = get_running_loop().create_task(self._cleanup())

    def __del__(self: "PingSocket") -> None:
        """
        Perform cleanup on delete.

        * Cancel expiration task.
        * Remove socket reader.
        """
        if not self.__force_del:
            return
        try:
            # Unsubscribe reader
            # get_running_loop() may raise Runtime Error
            get_running_loop().remove_reader(self.__sock_fd)
            # Stop cleanup task
            if self.__cleanup_task is not None:
                self.__cleanup_task.cancel()
        except RuntimeError:  # pragma: no cover
            pass  # Loop is already closed

    def clean_ip(self: "PingSocket", addr: str) -> str:
        """
        Normalize IP address to a stable form.

        Args:
            addr: IP address

        Returns:
            Normalized address
        """
        return self.__sock.clean_ip(addr)

    async def ping(
        self: "PingSocket",
        addr: str,
        size: Optional[int] = None,
        request_id: int = 0,
        seq: int = 0,
    ) -> Optional[float]:
        """
        Send ICMP echo request and await for result.

        Args:
            addr: Socket to ping.
            size: Packet size in bytes, including IP header.
            request_id: ICMP request id.
            seq: ICMP sequental number.
        """
        if ":" in addr:
            # Convert IPv6 address to compact form
            addr = self.__sock.clean_ip(addr)
        sid = f"{addr}-{request_id}-{seq}"
        fut: Future[Optional[float]] = get_running_loop().create_future()
        # Build and send the packet
        try:
            self.__sock.send(addr, request_id, seq, size or self.__size)
        except OSError:
            # Some kernels raise OSError (Network Unreachable)
            # when cannot find the route. Treat them as losses.
            fut.set_result(None)
            return await fut
        # Install future in the sessions
        self.__sessions[sid] = fut
        # Await response or timeout
        return await fut

    def _on_read(self: "PingSocket") -> None:
        """Handle socket read event."""
        # Get bulk read info from Rust side
        seen = self.__sock.recv()
        if seen is None:
            return
        # seen is the dict of sid -> rtt
        for sid, rtt in seen.items():
            # Find and pop the future in single call
            fut = self.__sessions.pop(sid, None)
            if fut:
                # Pass rtt to the future, unblock await in `ping`
                fut.set_result(float(rtt) / NS)

    async def _cleanup(self: "PingSocket") -> None:
        """Check for expired sessions and close them."""
        while True:
            # Wait for next cycle
            await sleep(self.__timeout)
            # Get a list of exired sids
            expired = self.__sock.get_expired()
            if not expired:
                continue
            # Iterate over expired sids
            for sid in expired:
                # Find and pop the future by single call
                fut = self.__sessions.pop(sid, None)
                if fut:
                    # Pass None to indicate the timeout
                    fut.set_result(None)



# ---------------------------------------------------------------------
# Gufo Ping: Ping implementation
# ---------------------------------------------------------------------
# Copyright (C) 2022-25, Gufo Labs
# ---------------------------------------------------------------------

"""Ping client implementation."""

# Python modules
import asyncio
import itertools
import random
from time import perf_counter
from typing import AsyncIterable, Dict, Iterable, Optional, Tuple, Union

# Gufo Labs modules
from .socket import IPv4, IPv6, PingSocket


class Ping(object):
    """
    High-performance asyncronous ICMPv4/ICMPv6 ping client.

    Args:
        size: Set outgoing packet's size, including IP header.
        src_addr: Set source address for outgoing packets.
            Depends upon address family. May be one of:
            * None - detect source address automatically.
            * str - containing source address for one address family.
            * Iterable of strings, containing multiple addresses
                which to be distributed among the address families.
                First address for given address family will be used.
        ttl: Set outgoing packet's TTL.
            Use OS defaults when empty.

            !!! warning

                This option is ignored on IPv6 socket due to
                issue [#4](https://github.com/gufolabs/gufo_ping/issues/4)
        tos: Set DSCP/TOS/TCLASS field to outgoing packets.
            Use OS defaults when empty.

            !!! warning

                This option is ignored on IPv6 socket due to
                issue [#2](https://github.com/gufolabs/gufo_ping/issues/2)
        timeout: Default timeout in seconds.
        send_buffer_size: Send buffer size.
            Use OS defaults when empty.
        recv_buffer_size: Receive buffer size.
            Use OS defaults when empty.
        coarse: Use CLOCK_MONOTONIC_COARSE when set,
            fall back to CLOCK_MONOTONIC otherwise.
        accelerated: Enable platform-dependend accelerated
            socket processing.

    Note:
        Opening the Raw Socket may require super-user priveleges
        or additional permissions. Refer to the operation system's
        documentation for details.

    Example:
        Ping single packet.

        ``` py
        from gufo.ping import Ping

        async def ping(address):
            p = Ping()
            rtt = await p.ping(address)
            print(rtt)
        ```

        Ping multiple packets.

        ``` py
        from gufo.ping import Ping

        async def ping(address):
            p = Ping()
            async for rtt in p.iter_rtt(address):
                print(rtt)
        ```
    """

    request_id = itertools.count(random.randint(0, 0xFFFF))

    def __init__(
        self: "Ping",
        size: int = 64,
        src_addr: Union[None, str, Iterable[str]] = None,
        ttl: Optional[int] = None,
        tos: Optional[int] = None,
        timeout: float = 1.0,
        send_buffer_size: Optional[int] = None,
        recv_buffer_size: Optional[int] = None,
        coarse: bool = False,
        accelerated: bool = True,
    ) -> None:
        self.__size = size
        self.__src_addr = self._get_src_addr(src_addr)
        self.__ttl = ttl
        self.__tos = tos
        self.__timeout = timeout
        self.__send_buffer_size = send_buffer_size
        self.__recv_buffer_size = recv_buffer_size
        self.__coarse = coarse
        self.__accelerated = accelerated
        self.__sockets: Dict[int, PingSocket] = {}

    @staticmethod
    def _get_afi(address: str) -> int:
        """
        Get address family (AFI) for a given address.

        Args:
            address: Address to ping.

        Returns:
            * `4` for IPv4
            * `6` for IPv6
        """
        if ":" in address:
            return IPv6
        return IPv4

    @staticmethod
    def _get_src_addr(addr: Union[None, str, Iterable[str]]) -> Dict[int, str]:
        """
        Parse source addresses.

        Parse source addresses and distribute them around address families.

        Args:
            addr: One of:
                * None - detect source address automatically.
                * str - containing source address for one address family.
                * Iterable of strings, containing multiple addresses
                    which to be distributed among the address families.
                    First address for given address family will be used.

        Returns:
            Dict of `address family` -> `source address`.
        """
        if not addr:
            return {}
        if isinstance(addr, str):
            return {Ping._get_afi(addr): addr}
        r: Dict[int, str] = {}
        for a in addr:
            afi = Ping._get_afi(a)
            if afi not in r:
                r[afi] = a
        return r

    def __get_socket(self: "Ping", address: str) -> PingSocket:
        """
        Get PingSocket instace.

        Get ping socket instance for specified address.
        Initialize when necessary.

        Args:
            address: Address to ping

        Returns:
            Initialized socket instance
        """
        afi = self._get_afi(address)
        sock = self.__sockets.get(afi)
        if not sock:
            sock = PingSocket(
                afi=afi,
                size=self.__size,
                src_addr=self.__src_addr.get(afi),
                ttl=self.__ttl,
                tos=self.__tos,
                timeout=self.__timeout,
                send_buffer_size=self.__send_buffer_size,
                recv_buffer_size=self.__recv_buffer_size,
                coarse=self.__coarse,
                accelerated=self.__accelerated,
            )
            self.__sockets[afi] = sock
        return sock

    def __get_request_id(self: "Ping") -> Tuple[int, int]:
        """
        Get request id.

        Generate ICMP request id and starting
        sequence number.

        Returns:
            Tuple of (`request_id`, `sequence`)
        """
        request_id = next(self.request_id) & 0xFFFF
        seq = random.randint(0, 0xFFFF)
        return request_id, seq

    async def ping(
        self: "Ping",
        addr: str,
        size: Optional[int] = None,
    ) -> Optional[float]:
        """
        Do ping probe.

        Send ICMP echo request to the given address and await
        for response or timeout.

        Args:
            addr: IPv4/IPv6 address to ping.
            size: Packet's size, including IP headers. Use PingSocket
                intialized defaults, when empty.

        Returns:
            * Round-trip time in seconds (as float) if success.
            * None - if failed or timed out.
        """
        sock = self.__get_socket(addr)
        request_id, seq = self.__get_request_id()
        return await sock.ping(addr, size=size, request_id=request_id, seq=seq)

    async def iter_rtt(
        self: "Ping",
        addr: str,
        *,
        size: Optional[int] = None,
        interval: Optional[float] = 1.0,
        count: Optional[int] = None,
    ) -> AsyncIterable[Optional[float]]:
        """
        Do the serie of ping probes.

        Send echo request every `interval` seconds,
        await and yield the result.

        Args:
            addr: Address to ping.
            size: Packets' size, including IP headers. Use PingSocket
                intialized defaults, when empty.
            interval: Interval between requests, in seconds.
            count: Stop after `count` requests, if set. Do not stop
                otherwise.

        Returns:
            Yields for each attempt:

            * Round-trip time in seconds (as float) if success.
            * None - if failed or timed out.

        """
        sock = self.__get_socket(addr)
        request_id, seq = self.__get_request_id()
        t0 = 0.0
        n = 0
        while True:
            if interval:
                t0 = perf_counter()
            yield await sock.ping(
                addr, size=size, request_id=request_id, seq=seq
            )
            seq = (seq + 1) & 0xFFFF
            if interval:
                dt = perf_counter() - t0
                if dt < interval:
                    await asyncio.sleep(interval - dt)
            n += 1
            if count and n >= count:
                break



# ---------------------------------------------------------------------
# Gufo Labs: CI test
# ---------------------------------------------------------------------
# Copyright (C) 2022-23, Gufo Labs
# See LICENSE.md for details
# ---------------------------------------------------------------------

# Python modules
import inspect
import os
import sys
from dataclasses import dataclass
from typing import Iterable

# Third-party modules
import pytest
import yaml


def _get_root() -> str:
    mod_path = inspect.getfile(sys.modules[__name__])
    rel_root = os.path.dirname(mod_path)
    return os.path.abspath(os.path.join(rel_root, ".."))


VERSIONS = [
    "actions/cache@v4",
    "actions/checkout@v4",
    "actions/setup-python@v5",
    "actions/download-artifact@v4",
    "actions/upload-artifact@v4",
    "pypa/gh-action-pypi-publish@release/v1",
]


@dataclass
class Action(object):
    path: str
    job: str
    step: str
    action: str
    version: str
    expected: str


def action_label(action: Action) -> str:
    return f"{action.path}: {action.job}/{action.step}: {action.action}"


def _iter_actions() -> Iterable[Action]:
    versions = {a.split("@")[0]: a.split("@")[1] for a in VERSIONS}
    root = os.path.join(_get_root(), ".github", "workflows")
    for fn in os.listdir(root):
        if fn.startswith(".") or not fn.endswith(".yml"):
            continue
        path = os.path.join(root, fn)
        with open(path) as f:
            data = yaml.safe_load(f)
        for job in data["jobs"]:
            for step in data["jobs"][job]["steps"]:
                if "uses" in step:
                    uses = step["uses"]
                    for v in versions:
                        if uses.startswith(f"{v}@"):
                            yield Action(
                                path=fn,
                                job=job,
                                step=step["name"],
                                action=v,
                                version=uses.split("@")[1],
                                expected=versions[v],
                            )
                            break


@pytest.mark.parametrize("action", list(_iter_actions()), ids=action_label)
def test_actions(action: Action) -> None:
    loc = f"{action.path}: {action.job}/{action.step}"
    v_exp = f"{action.action}@{action.expected}"
    msg = f"{loc}: {v_exp} required (@{action.version} used)"
    assert action.version == action.expected, msg



# ---------------------------------------------------------------------
# Gufo Labs: Project structure tests
# ---------------------------------------------------------------------
# Copyright (C) 2022-24, Gufo Labs
# See LICENSE.md for details
# ---------------------------------------------------------------------

# Python modules
import inspect
import os
import sys
from pathlib import Path
from typing import Tuple, Union

# Third-party modules
import pytest


def _get_root() -> str:
    mod_path = inspect.getfile(sys.modules[__name__])
    rel_root = os.path.dirname(mod_path)
    return os.path.abspath(os.path.join(rel_root, ".."))


def _get_project_info() -> Tuple[str, str]:
    """
    Get project information.

    Returns:
        Tuple of (project path, project module)
    """

    def explore_dir(*args: str) -> str:
        def is_not_rust(path: Path) -> bool:
            if not path.is_dir():
                return True
            return not any(path.rglob("*.rs"))

        root = Path(*args)
        d = [
            f
            for f in os.listdir(os.path.join(*args))
            if not f.startswith(".")
            and not f.startswith("_")
            and not f.endswith(".egg-info")
            and not f.endswith(".rs")
        ]
        if len(d) > 1:
            d = [x for x in d if is_not_rust(root / x)]
        assert len(d) == 1
        return d[0]

    ns = explore_dir(ROOT, "src")
    if ns == "gufo":
        # gufo.* namespace
        pkg = explore_dir(ROOT, "src", ns)
        return os.path.join("src", ns, pkg), f"{ns}.{pkg}"
    return os.path.join("src", ns), ns


ROOT = _get_root()
PROJECT_SRC, PROJECT_MODULE = _get_project_info()

REQUIRED_FILES = [
    ".devcontainer/devcontainer.json",
    ".github/CODEOWNERS",
    ".github/ISSUE_TEMPLATE/bug-report.yml",
    ".github/ISSUE_TEMPLATE/feature-request.yml",
    ".github/PULL_REQUEST_TEMPLATE.md",
    ".gitignore",
    ".requirements/docs.txt",
    ".requirements/lint.txt",
    ".requirements/test.txt",
    "CHANGELOG.md",
    "CITATION.cff",
    "CODE_OF_CONDUCT.md",
    "CONTRIBUTING.md",
    "Dockerfile",
    "LICENSE.md",
    "README.md",
    "SECURITY.md",
    "docs/assets/logo.png",
    "docs/dev/codebase.md",
    "docs/dev/codequality.md",
    "docs/dev/common.md",
    "docs/dev/environment.md",
    "docs/dev/testing.md",
    "docs/faq.md",
    "docs/index.md",
    ("docs/installation.md", "docs/installation/index.md"),
    "mkdocs.yml",
    "pyproject.toml",
    f"{PROJECT_SRC}/__init__.py",
    f"{PROJECT_SRC}/py.typed",
    "tests/test_docs.py",
    "tests/test_project.py",
]


def test_required_is_sorted() -> None:
    def q(name: Union[str, Tuple[str, ...]]) -> Tuple[str, ...]:
        if isinstance(name, str):
            return (name,)
        return name

    normalized = [q(x) for x in REQUIRED_FILES]
    assert sorted(normalized) == normalized, "REQUIRED_FILES must be sorted"


@pytest.mark.parametrize("name", REQUIRED_FILES)
def test_required_files(name: Union[str, Tuple[str, ...]]) -> None:
    if isinstance(name, str):
        full_path = os.path.join(ROOT, name)
        assert os.path.exists(full_path), f"File {name} is missed"
    else:
        full_paths = [os.path.join(ROOT, n) for n in name]
        present = any(os.path.exists(n) for n in full_paths)
        assert present, f"Any of files {', '.join(full_paths)} must be exist"


def test_version() -> None:
    m = __import__(PROJECT_MODULE, {}, {}, "*")
    assert hasattr(m, "__version__"), "__init__.py must contain __version__"



# ---------------------------------------------------------------------
# Gufo Ping: ICMPv4/ICMPv6 ping implementation
# Python build
# ---------------------------------------------------------------------
# Copyright (C) 2022-23, Gufo Labs
# ---------------------------------------------------------------------

# Python modules
from typing import Optional, Sequence
import os

# Third-party modules
from setuptools import setup
from setuptools_rust import RustExtension


def _from_env(name: str) -> Optional[Sequence[str]]:
    v = os.environ.get(name, "").strip()
    if not v:
        return None
    return v.split()


def get_rustc_flags() -> Optional[Sequence[str]]:
    return _from_env("BUILD_RUSTC_FLAGS")


def get_cargo_flags() -> Optional[Sequence[str]]:
    return _from_env("BUILD_CARGO_FLAGS")


setup(
    rust_extensions=[
        RustExtension(
            "gufo.ping._fast",
            args=get_cargo_flags(),
            rustc_flags=get_rustc_flags(),
        )
    ],
)



# ----------------------------------------------------------------------
# Gufo Traceroute: docs tests
# ----------------------------------------------------------------------
# Copyright (C) 2022-23, Gufo Labs
# See LICENSE.md for details
# ----------------------------------------------------------------------

# Python modules
import os
import re
from functools import lru_cache
from typing import List, Optional, Set

# Third-party modules
import pytest

_doc_files: Optional[List[str]] = None

rx_link = re.compile(r"\[([^\]]+)\]\[([^\]]+)\]", re.MULTILINE)
rx_link_def = re.compile(r"^\[([^\]]+)\]:", re.MULTILINE)
rx_footnote = re.compile(r"[^\]]\[(\^\d+)\][^\[]", re.MULTILINE)
rx_datatracker_ietf = re.compile(
    r"https://datatracker\.ietf\.org/doc/[^/]+/(rfc\d+)", re.MULTILINE
)


@lru_cache(maxsize=1)
def get_docs() -> List[str]:
    doc_files: List[str] = []
    for root, _, files in os.walk("docs"):
        for f in files:
            if f.endswith(".md") and not f.startswith("."):
                doc_files.append(os.path.join(root, f))
    return doc_files


def get_file(path: str) -> str:
    with open(path) as f:
        return f.read()


@pytest.mark.parametrize("doc", get_docs())
def test_links(doc: str) -> None:
    data = get_file(doc)
    links: Set[str] = set()
    defs: Set[str] = set()
    for match in rx_link.finditer(data):
        links.add(match.group(2))
    for match in rx_footnote.finditer(data):
        links.add(match.group(1))
    for match in rx_link_def.finditer(data):
        d = match.group(1)
        assert d not in defs, f"Link already defined: {d}"
        assert d in links, f"Unused link definition: {d}"


@pytest.mark.parametrize("doc", get_docs())
def test_rfc_links(doc: str) -> None:
    data = get_file(doc)
    match = rx_datatracker_ietf.search(data)
    assert not match, f"{match.group(0)} link used. Must be https://www.rfc-editor.org/rfc/{match.group(1)}.html"



# ---------------------------------------------------------------------
# Gufo Ping: SocketProto
# ---------------------------------------------------------------------
# Copyright (C) 2022-23, Gufo Labs
# ---------------------------------------------------------------------

"""SocketWrapper protocol definition."""

# Python modules
from typing import Dict, List, Optional, Protocol


class SocketProto(Protocol):
    """
    SocketWrapper protocol.

    Socket wrapper is the [PyO3](https://pyo3.rs)-wrapped Rust code,
    implementing low-level details of the PingSocket.
    """

    def __init__(self: "SocketProto", afi: int) -> None: ...

    def set_timeout(self: "SocketProto", timeout: int) -> None:
        """
        Set default ping timeout.

        Args:
            timeout: Ping timeout, in nanoseconds.
        """
        ...

    def bind(self: "SocketProto", addr: str) -> None:
        """
        Bind to source address.

        Args:
            addr: Source Address
        """

    def set_ttl(self: "SocketProto", ttl: int) -> None:
        """
        Change outgoing packets' time-to-live field.

        Args:
            ttl: TTL value.
        """
        ...

    def set_unicast_hops(self: "SocketProto", ttl: int) -> None:
        """
        Change outgoing packets' unicast hops (IPv6 TTL).

        Args:
            ttl: TTL value.
        """
        ...

    def set_tos(self: "SocketProto", tos: int) -> None:
        """
        Change outgoing packets' ToS/DSCP field.

        Args:
            tos: ToS value.
        """
        ...

    def set_tclass(self: "SocketProto", tclass: int) -> None:
        """
        Change outgoing packets' IPv6 TCLASS field.

        Args:
            tclass: TCLASS value.
        """
        ...

    def set_coarse(self: "SocketProto", ct: bool) -> None:
        """
        Switch between the internal timer implemenetation.

        Args:
            ct: Use

                * `CLOCK_MONOTONIC_COARSE` if True
                * `CLOCK_MONOTONIC` if False
        """
        ...

    def set_send_buffer_size(self: "SocketProto", size: int) -> None:
        """
        Set outgoing socket's buffer size.

        If the requested
        size is too big, adjust to proper size.

        Args:
            size: Requested send buffer size, in bytes.
        """
        ...

    def set_recv_buffer_size(self: "SocketProto", size: int) -> None:
        """
        Set incoming socket's buffer size.

        If the requested
        size is too big, adjust to proper size.

        Args:
            size: Requested recv buffer size, in bytes.
        """
        ...

    def set_accelerated(self: "SocketProto", a: bool) -> None:
        """
        Enable platform-dependend raw socket processing.

        Args:
            a: Enable or disable an acceleration.
                * True - enable, when platform supports acceleration.
                * False - disable the acceleration.
        """

    def get_fd(
        self: "SocketProto",
    ) -> int:  # @todo: Shold be FileDescriptorLike
        """
        Get socket's file descriptor.

        Returns:
            file descriptor for open socket.
        """
        ...

    def send(
        self: "SocketProto", addr: str, request_id: int, seq: int, size: int
    ) -> None:
        """
        Generate and send icmp request packet.

        Args:
            addr: Destination address.
            request_id: ICMP request id.
            seq: ICMP sequental number.
            size: Outgoing packet's size in bytes, including IP header.
        """
        ...

    def recv(self: "SocketProto") -> Optional[Dict[str, float]]:
        """
        Receive all awaiting packets.

        Returns:
            * `None` - when no packets received.
            * Dict of `session id` -> `rtt`,
                where `session id` is the string of
                <address>-<request_id>-<seq>,
                and `rtt` - is the measured round-trip-time in nanoseconds.
        """
        ...

    def get_expired(self: "SocketProto") -> Optional[List[str]]:
        """
        Get list of sessions with expired timeouts.

        Returns:
            * `None` - when no sessions expired.
            * List of expired sessionn ids, where each session id
                has the format: <address>-<request_id>-<seq>
        """
        ...

    def clean_ip(self: "SocketProto", addr: str) -> str:
        """
        Normalize IP address to a stable form.

        Args:
            addr: IP address

        Returns:
            Normalized address
        """
        ...
